from django_pds.conf import settings
from django_pds.core.managers import GenericUpdateCommandManager
from django_pds.core.utils import print_traceback

SECURITY_ATTRIBUTES = settings.SECURITY_ATTRIBUTES
READ_ONLY_FIELDS = settings.READ_ONLY_FIELDS


def data_update(document_name, data, user_id=None, ignore_security=False, track_error=False):
    results = []

    data_array = []
    if isinstance(data, dict):
        data_array.append(data)
    if isinstance(data, list):
        data_array = data

    update_manager = GenericUpdateCommandManager()

    for d in data_array:

        try:

            if ignore_security:
                error, res = update_manager.update_one(document_name, d)
                results.append({
                    "error": error,
                    "message": res
                })
                continue

            data_fields = set(d.keys())
            security_attributes = set(SECURITY_ATTRIBUTES)
            common_fields = update_manager.common_fields(data_fields, security_attributes)

            if len(common_fields) > 0:
                fields = ','.join(common_fields)
                results.append({
                    "error": True,
                    "message": f'following security attributes {fields} found in the json data'
                })
                continue

            read_only_fields = set(READ_ONLY_FIELDS)
            rof = update_manager.common_fields(data_fields, read_only_fields)
            if len(rof) > 0:
                fields = ','.join(rof)
                results.append({
                    "error": True,
                    "message": f'the following read only attributes {fields} found in the json data'
                })
                continue

            can_update = update_manager.can_update(document_name, d.get('ItemId', None), user_id)

            if can_update:
                err, res = update_manager.update_one(document_name, d, user_id)
                results.append({
                    "error": err,
                    "message": res
                })
            else:
                results.append({
                    "error": True,
                    "message": "access denied, you don't have sufficient permission to update"
                })

        except BaseException as e:

            if track_error:
                print_traceback()

            results.append({
                "error": True,
                "message": str(e)
            })



import os
import traceback

from django.core.exceptions import ImproperlyConfigured
from mongoengine import base


def is_abstract_document(document_name):
    try:
        __doc = get_document(document_name)
        return __doc._meta.get('abstract', False)
    except BaseException as e:
        return False


def get_document(document_name):
    try:
        return base.get_document(document_name)
    except BaseException as e:
        return None


def get_fields(document_name, with_auto_id=False):
    __doc = get_document(document_name)
    _fields = list(__doc._fields.keys())
    if not with_auto_id:
        elem = 'auto_id_0'
        if elem in _fields:
            _fields.remove(elem)
    return _fields


def origin(request):
    if request.META is not None:
        return request.META.get('HTTP_ORIGIN')
    return None


def get_host(request):
    host = origin(request)
    if host is None:
        return True, None
    replace = ['https://', 'http://', 'www.', '/']
    for item in replace:
        host = host.replace(item, '')
    hosts = host.split(':')
    return hosts[0]


def authorization_token(request):
    return request.META.get('HTTP_AUTHORIZATION', None)


def get_content_type(request):
    return request.META.get('CONTENT_TYPE', None)


def path(request):
    return str(request.META.get('PATH_INFO', ''))


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def print_traceback():
    tb = traceback.format_exc()
    print(tb)


def get_environment(key, raise_exception=True, default=None):
    """ Get the environment variable or return exception """

    try:
        return os.environ[key]
    except KeyError:
        if raise_exception:
            raise ImproperlyConfigured
        else:
            return default



from mongoengine import StringField, ListField

from django_pds.core.base import SimpleBaseDocument


class UserRoleMap(SimpleBaseDocument):
    RoleName = StringField(required=True)
    IdsAllowedToRead = ListField(StringField(required=True), default=[])
    meta = {'collection': 'UserRoleMaps'}



ORDER_BY = "orderby"
FROM = "from"
FILTER = "filter"
WHERE = 'where'
SELECT = 'select'
PAGE_SIZE = "pagesize"
PAGE_NUM = "pagenumber"
EQUAL = '='
LT = '<'
GT = '>'
AND = '&'
OR = '@'
EMPTY = ""
COMMA = ','
PARENTHESIS_LEFT = '('
PARENTHESIS_RIGHT = ')'
RAW_WHERE = 'RAW_WHERE'



from django_pds.core.managers import GenericInsertCommandManager
from .update import data_update
from .write import data_insert


def data_upsert(document_name, data, user_id=None, ignore_security=False, force_upsert=False, track_error=False):
    try:
        insert_ctrl = GenericInsertCommandManager()
        already_exists = insert_ctrl.already_exists(document_name, data.get('ItemId', None))

        if not already_exists:
            return data_insert(user_id, document_name, data, ignore_security, force_upsert, track_error)
        else:
            return data_update(user_id, document_name, data, ignore_security, track_error)

    except BaseException as e:
        return True, str(e)



from django.middleware.http import MiddlewareMixin
from django.urls import resolve
from django.urls.resolvers import Resolver404

from django_pds.core.rest.exceptions import url_not_found


class UrlPathExistsMiddleware(MiddlewareMixin):

    def process_request(self, request):
        try:
            resolve(request.path)
            return None
        except Resolver404:
            return url_not_found(request)



def error_response(message=None, status=None):
    return {
        "error_status": status or 400,
        "error": "bad_request",
        "error_description": message or "The request is not valid."
    }


def success_response(data=None, message=None, status=None):
    return {
        "success_status": status or 200,
        "error": None,
        "success_description": message or "request valid",
        "responsible": data or None
    }


def error_response_read_only_fields(fields, message=None, status=None):
    return {
        "error_status": status or 400,
        "error": "bad_request",
        "error_description": message or "read only fields found in json string",
        "responsible": fields
    }


def success_response_with_total_records(data, count, message=None, status=None):
    return {
        "success_status": status or 200,
        "error": None,
        "success_description": message or "request valid",
        "results": data,
        "total_records": count
    }



__all__ = [
    'InsertPermissionError',
    'UpdatePermissionError',
    'ReadPermissionError',
    'DeletePermissionError'
]


class PdsPermissionError(Exception):
    pass


class InsertPermissionError(PdsPermissionError):
    pass


class ReadPermissionError(PdsPermissionError):
    pass


class DeletePermissionError(PdsPermissionError):
    pass


class UpdatePermissionError(PdsPermissionError):
    pass



from django_pds.conf import settings
from django_pds.core.managers import DefaultPermissionSettingsManager, GenericInsertCommandManager
from django_pds.core.utils import print_traceback

security_attributes = settings.SECURITY_ATTRIBUTES
read_only_fields = settings.READ_ONLY_FIELDS


def data_insert(document_name, data, user_id=None, ignore_security=False, force_insert=False, track_error=False):
    results = []
    insert_manager = GenericInsertCommandManager()

    data_array = []

    if isinstance(data, dict):
        data_array.append(data)
    elif isinstance(data, list):
        data_array = data

    for d in data_array:

        try:

            if force_insert:
                already_exists = insert_manager.already_exists(document_name, d.get('ItemId', None))
                if already_exists and not force_insert:
                    message = "document ItemId already exists, you can't create new collection " \
                              "with the same ItemId, if you want to update, " \
                              "use pds update or upsert method"
                    results.append({
                        'error': True,
                        'message': message
                    })
                    continue

            if ignore_security:
                error, response = insert_manager.insert_one(document_name, d)
                results.append({
                    "error": error,
                    "message": response
                })
                continue

            entity_permission = DefaultPermissionSettingsManager()
            can_insert = entity_permission.can_insert(document_name, user_id)

            if can_insert:

                data_fields = set(d.keys())
                rof = set(read_only_fields)
                common_fields = data_fields.intersection(rof)
                if len(common_fields) > 0:
                    fields = ','.join(common_fields)
                    results.append({
                        "error": True,
                        "message": f'the following read only attributes {fields} found in the json data'
                    })
                    continue
                sec_attr = set(security_attributes)
                common_fields = data_fields.intersection(sec_attr)
                if len(common_fields) > 0:
                    fields = ','.join(common_fields)
                    results.append({
                        "error": True,
                        "message": f'following security attributes {fields} found in the json data'
                    })
                    continue
                # insert data after all the verification and security checking's
                error, permissions = entity_permission.get_document_name_permissions(document_name)
                if error:
                    results.append({
                        "error": error,
                        "message": permissions
                    })
                    continue
                err, response = insert_manager.insert_one(document_name, d, user_id, permissions)
                results.append({
                    "error": err,
                    "message": response
                })
            else:
                results.append({
                    "error": True,
                    "message": "access denied, you don't have sufficient permission to insert data"
                })
        except BaseException as e:
            if track_error:
                print_traceback()
            results.append({
                "error": True,
                "message": str(e)
            })



from rest_framework.views import APIView

from .exceptions import method_not_allowed


class BaseAPIView(APIView):

    def http_method_not_allowed(self, request, *args, **kwargs):
        return method_not_allowed()



from .manager import BaseManager


class GenericDeleteCommandManager(BaseManager):

    def has_permission(self, document_name, document_id, user_id):
        try:
            Model = self.get_document(document_name)
            mod = Model.objects(ItemId=document_id, IdsAllowedToDelete=user_id)
            return mod.count() > 0
        except BaseException as e:
            return False

    def delete(self, document_name, document_id):

        base_instance = self.is_base_instance(document_name)
        simple_base_instance = self.is_simple_base_doc_instance(document_name)

        if not base_instance and not simple_base_instance:
            return True, 'Document type must be `BaseDocument` ' \
                         'or `SimpleBaseDocument` ' \
                         'from django_pds.core.base Module'

        try:
            Model = self.get_document(document_name)
            mod = Model.objects(ItemId=document_id)
            mod.delete()
            return True, "Delete success!"
        except BaseException as e:
            return False, str(e)

    def permission_and_delete(self, entity, user_id, item_id):
        if self.has_permission(entity, user_id, item_id):
            return self.delete(entity, item_id)
        return False



from django.utils.timezone import now
from mongoengine import Document, StringField, ListField, DateTimeField


class SimpleBaseDocument(Document):
    ItemId = StringField(required=True, max_length=36, db_field='_id')
    meta = {
        'allow_inheritance': False,
        'abstract': True,
        'strict': True
    }

    def __str__(self):
        return self.ItemId


class BaseDocument(SimpleBaseDocument):
    CreatedBy = StringField(required=False, max_length=36)
    CreateDate = DateTimeField(default=now)

    Language = StringField(required=True, default='en-US')

    LastUpdateDate = DateTimeField(default=now)
    LastUpdateBy = StringField(required=False)

    Tags = ListField(StringField(min_length=3), default=[])

    IdsAllowedToRead = ListField(StringField(max_length=36), default=[])
    IdsAllowedToWrite = ListField(StringField(max_length=36), default=[])
    IdsAllowedToUpdate = ListField(StringField(max_length=36), default=[])
    IdsAllowedToDelete = ListField(StringField(max_length=36), default=[])

    RolesAllowedToRead = ListField(StringField(max_length=36), default=[])
    RolesAllowedToWrite = ListField(StringField(max_length=36), default=[])
    RolesAllowedToUpdate = ListField(StringField(max_length=36), default=[])
    RolesAllowedToDelete = ListField(StringField(max_length=36), default=[])

    meta = {
        'allow_inheritance': False,
        'abstract': True,
        'strict': True
    }

    @classmethod
    def pre_save(cls, sender, document, **kwargs):
        document.CreateDate = now()
        document.LastUpdateDate = now()


from django_pds.core.base import BaseDocument, SimpleBaseDocument
from django_pds.core.utils import get_document as document_provider


class BaseManager:

    def common_fields(self, a, b):
        return a.intersection(b)

    def get_document(self, document_name):
        return document_provider(document_name)

    def is_base_instance(self, document_name):
        doc = self.get_document(document_name)
        return issubclass(doc, BaseDocument)

    def is_simple_base_doc_instance(self, document_name):
        doc = self.get_document(document_name)
        return issubclass(doc, SimpleBaseDocument)



from mongoengine import QuerySet


class PdsBaseQuerySet(QuerySet):

    def __has_delete_permission(self, document_id, user_id):
        docs = self.filter(ItemId=document_id, IdsAllowedToDelete=user_id)
        return docs.count() > 0

    def __pds_delete(self, document_id):
        docs = self.filter(ItemId=document_id)
        count = docs.delete()
        return count > 0

    def pds_delete(self, document_id, user_id):
        if self.__has_delete_permission(document_id, user_id):
            return self.__pds_delete(document_id)
        return False

    def exists(self, document_id):
        return self.filter(ItemId=document_id).count() > 0



from .delete import GenericDeleteCommandManager
from .entity import DefaultPermissionSettingsManager
from .insert import GenericInsertCommandManager
from .kwargs import KwargsManager
from .manager import BaseManager
from .read import GenericReadManager
from .required import RequiredManager
from .update import GenericUpdateCommandManager
from .userreadabledata import UserReadableDataManager
from .userrolemaps import UserRoleMapsManager

__all__ = [
    'GenericDeleteCommandManager',
    'DefaultPermissionSettingsManager',
    'GenericInsertCommandManager',
    'KwargsManager',
    'BaseManager',
    'GenericReadManager',
    'RequiredManager',
    'GenericUpdateCommandManager',
    'UserReadableDataManager',
    'UserRoleMapsManager'
]



from .manager import BaseManager


class GenericReadManager(BaseManager):

    def read(self, document_name, q, page_size, page_number, order_by=None):

        if order_by is None:
            order_by = []

        if page_size > 100:
            page_size = 100
        if page_number < 1:
            page_number = 0
        else:
            page_number = page_number - 1

        try:
            data = self.get_document(document_name).objects.filter(q).order_by(*order_by).skip(
                page_size * page_number).limit(page_size)
            cnt = data.count()
            return data, cnt
        except:
            return [], 0

    def read_by_id(self, model, item_id):
        try:
            data = model.objects(ItemId=item_id)
            cnt = data.count()
            return data, cnt
        except:
            return [], 0



from uuid import uuid4

from .manager import BaseManager

USER_ROLE_MAP = 'UserRoleMap'


class UserRoleMapsManager(BaseManager):

    def get_user_roles(self, user_id):
        return self.get_document(USER_ROLE_MAP).objects(IdsAllowedToRead=user_id).distinct(field='RoleName')

    def create_user_role_map(self, role_name, user_id):
        try:
            role = self.get_document(USER_ROLE_MAP)()
            role.ItemId = str(uuid4())
            role.RoleName = role_name
            role.IdsAllowedToRead = [user_id]
            role.save()
            return False, role.ItemId
        except BaseException as e:
            return True, e



from django_pds.conf import settings
from django_pds.core.settings import SECURITY_ATTRIBUTES
from django_pds.serializers import GenericSerializerAlpha
from .manager import BaseManager
from .userrolemaps import UserRoleMapsManager

ENTITY_DEFAULT_PERMISSION_SETTINGS = settings.DOCUMENT_ENTITY_DEFAULT_PERMISSION_SETTING
ADMIN = 'admin'


class DefaultPermissionSettingsManager(BaseManager):

    def __has_permission(self, document_name, row, user_id):
        permissions = self.get_document(ENTITY_DEFAULT_PERMISSION_SETTINGS).objects(
            EntityName=document_name)
        if permissions.count() > 0:
            permission = permissions[0]
            user_role_manager = UserRoleMapsManager()
            user_roles = set(user_role_manager.get_user_roles(user_id))
            permitted_roles = set(getattr(permission, row, []))
            permitted_roles.add(ADMIN)
            common_roles = user_roles.intersection(permitted_roles)
            return len(common_roles) > 0
        return False

    def can_insert(self, document_name, user_id):
        return self.__has_permission(document_name, 'RolesAllowedToWrite', user_id)

    def can_delete(self, document_name, user_id):
        return self.__has_permission(document_name, 'RolesAllowedToDelete', user_id)

    def can_read(self, document_name, user_id):
        return self.__has_permission(document_name, 'RolesAllowedToRead', user_id)

    def can_update(self, document_name, user_id):
        return self.__has_permission(document_name, 'RolesAllowedToUpdate', user_id)

    def get_document_name_permissions(self, document_name):
        try:
            permissions = self.get_document(ENTITY_DEFAULT_PERMISSION_SETTINGS).objects(EntityName=document_name)
            if permissions.count() > 0:
                permission = permissions[0]
                json = GenericSerializerAlpha().document(ENTITY_DEFAULT_PERMISSION_SETTINGS).fields(
                    SECURITY_ATTRIBUTES).serialize(permission)
                return False, json.data
            return True, 'permission: Not Allowed'
        except BaseException as e:
            return True, str(e)



from mongoengine import StringField, ListField

from django_pds.core.base import SimpleBaseDocument


class UserReadableData(SimpleBaseDocument):
    EntityName = StringField(max_length=120, required=True, null=False)
    Role = StringField(max_length=120, required=False, default='default')
    UserReadableFields = ListField(StringField(null=False), required=True)

    meta = {'collection': 'UserReadableDatas'}



from rest_framework import status
from rest_framework.response import Response

from django_pds.core.rest.response import error_response


class RequiredManager:

    def required(self, request, *args):

        if request.META.get('REQUEST_METHOD') == 'GET':
            if len(args) > 0:
                response = error_response('Get method should not have required parameters. Config error.', status=400)
                return True, Response(response, status=status.HTTP_400_BAD_REQUEST)
        params, files = request.data, request.FILES
        required_fields = args
        missing_fields = []

        if len(required_fields) == 0:
            return False, None

        ignore_on_files = []

        for field in required_fields:
            if params.get(field) is None:
                missing_fields.append(field)
            else:
                ignore_on_files.append(field)

        for field in required_fields:
            if files.get(field) is None:
                if field not in ignore_on_files and field not in missing_fields:
                    missing_fields.append(field)

        if len(missing_fields) == 0:
            return False, None

        response = {
            "error": 400,
            "error_message": "missing required payload fields in this request",
            "missing_fields": missing_fields
        }

        return True, Response(response, status=status.HTTP_400_BAD_REQUEST)



from mongoengine import base
from mongoengine.queryset.queryset import QuerySet
from rest_framework_mongoengine import serializers


class GenericSerializerAlpha:
    __fields = ()
    __document = None

    def __init__(self, document_name=None, fields=None):
        if document_name:
            self.__document = base.get_document(document_name)
        if fields:
            self.__fields = fields

    def fields(self, fields='__all__'):
        self.__fields = fields
        return self

    def document(self, document_name):
        self.__document = base.get_document(document_name)
        return self

    def delete(self):
        self.__del__()

    def __del__(self):
        pass

    def select(self, field):
        if type(self.__fields) == tuple:
            self.__fields = self.__fields + (field,)
        else:
            self.__fields = (field,)
        return self

    def select_all(self):
        self.__fields = '__all__'
        return self

    def serialize(self, data, many=True, allow_null=True):

        if type(data) is not QuerySet:
            many = False

        resp = self._GenSerializerAlpha(data, many=many, allow_null=allow_null, fields=self.__fields,
                                        document=self.__document)
        return resp

    class _GenSerializerAlpha(serializers.DocumentSerializer):

        def __init__(self, *args, **kwargs):
            __fields = kwargs.pop('fields', '__all__')
            __document = kwargs.pop('document', None)
            super(GenericSerializerAlpha._GenSerializerAlpha, self).__init__(*args, **kwargs)
            self.Meta.fields = __fields
            self.Meta.model = __document

        class Meta:
            model = None
            fields = '__all__'
